import pytest

from openskill import Rating, create_rating, rate


def test_plackett_luce_series():
    p00 = Rating()
    p10 = Rating()
    p20 = Rating()
    p30 = Rating()
    p40 = Rating()

    result = rate([[p00], [p10], [p20], [p30], [p40]], score=[9, 7, 7, 5, 5])
    [[p01], [p11], [p21], [p31], [p41]] = result
    p01 = create_rating(p01)
    p11 = create_rating(p11)
    p21 = create_rating(p21)
    p31 = create_rating(p31)
    p41 = create_rating(p41)

    p02 = p01
    p32 = p31
    result = rate([[p41], [p21], [p11]], score=[9, 5, 5])
    [[p42], [p22], [p12]] = result
    p42 = create_rating(p42)
    p22 = create_rating(p22)
    p12 = create_rating(p12)

    p43 = p42
    result = rate([[p32], [p12], [p22], [p02]], score=[9, 9, 7, 7])
    [[p33], [p13], [p23], [p03]] = result
    p33 = create_rating(p33)
    p13 = create_rating(p13)
    p23 = create_rating(p23)
    p03 = create_rating(p03)

    assert p03.mu == pytest.approx(26.353761103)
    assert p03.sigma == pytest.approx(8.11102706)
    assert p13.mu == pytest.approx(24.618479789)
    assert p13.sigma == pytest.approx(7.90533551)
    assert p23.mu == pytest.approx(23.065819512)
    assert p23.sigma == pytest.approx(7.822005595)
    assert p33.mu == pytest.approx(24.476332403)
    assert p33.sigma == pytest.approx(8.106111471)
    assert p43.mu == pytest.approx(26.385499685)
    assert p43.sigma == pytest.approx(8.054090809)



from openskill.models.bradley_terry_full import BradleyTerryFull
from openskill.models.bradley_terry_part import BradleyTerryPart
from openskill.models.plackett_luce import PlackettLuce
from openskill.models.thurston_mosteller_full import ThurstonMostellerFull
from openskill.models.thurston_mosteller_part import ThurstonMostellerPart



import pytest

from openskill.statistics import vt


def test_vt():
    assert vt(-1000, -100) == 1100
    assert vt(1000, -100) == -1100
    assert vt(-1000, 1000) == pytest.approx(0.79788, 0.00001)
    assert vt(0, 1000) == 0



from openskill.util import ladder_pairs


def test_ladder_pairs():
    assert ladder_pairs([]) == [[]]
    assert ladder_pairs([1]) == [[]]
    assert ladder_pairs([1, 2]) == [[2], [1]]
    assert ladder_pairs([1, 2, 3]) == [[2], [1, 3], [2]]
    assert ladder_pairs([1, 2, 3, 4]) == [[2], [1, 3], [2, 4], [3]]



import pytest

from openskill import Rating, create_rating, rate
from openskill.models import ThurstonMostellerPart


def test_thurston_mosteller_part_series():
    p00 = Rating()
    p10 = Rating()
    p20 = Rating()
    p30 = Rating()
    p40 = Rating()

    result = rate(
        [[p00], [p10], [p20], [p30], [p40]],
        score=[9, 7, 7, 5, 5],
        epsilon=0.1,
        gamma=lambda *_: 1,
        model=ThurstonMostellerPart,
    )
    [[p01], [p11], [p21], [p31], [p41]] = result
    p01 = create_rating(p01)
    p11 = create_rating(p11)
    p21 = create_rating(p21)
    p31 = create_rating(p31)
    p41 = create_rating(p41)

    assert p01.mu == pytest.approx(27.108980741)
    assert p01.sigma == pytest.approx(8.063357519)
    assert p11.mu == pytest.approx(22.891019259)
    assert p11.sigma == pytest.approx(7.620583708)
    assert p21.mu == pytest.approx(27.108980741)
    assert p21.sigma == pytest.approx(7.620583708)
    assert p31.mu == pytest.approx(22.891019259)
    assert p31.sigma == pytest.approx(7.620583708)
    assert p41.mu == pytest.approx(25.0)
    assert p41.sigma == pytest.approx(7.905694531)



import pytest

from openskill.rating import Rating
from openskill.util import team_rating, util_c, util_sum_q

r = Rating()
team_1 = [r]
team_2 = [r, r]


def test_util_sum_q():
    # Compute Values
    ratings = team_rating([team_1, team_2])
    c = util_c(ratings)
    assert util_sum_q(ratings, c) == [29.67892702634643, 24.70819334370875]

    # 5 v 5
    ratings = team_rating([[r, r, r, r, r], [r, r, r, r, r]])
    c = util_c(ratings)
    sum_q = util_sum_q(ratings, c)
    assert sum_q[0] == pytest.approx(204.84, 0.0001)
    assert sum_q[1] == pytest.approx(102.42, 0.0001)



import math

from openskill.constants import Constants
from openskill.statistics import v, vt, w, wt
from openskill.util import gamma, ladder_pairs, team_rating


class ThurstonMostellerPart:
    def __init__(self, game, **options):
        self.constants = Constants(**options)
        self.EPSILON = self.constants.EPSILON
        self.TWO_BETA_SQUARED = self.constants.TWO_BETA_SQUARED
        self.team_ratings = team_rating(game, **options)
        self.gamma = gamma(**options)
        self.adjacent_teams = ladder_pairs(self.team_ratings)

    def calculate(self):
        def i_map(i_team_rating, i_adjacents):
            i_mu, i_sigma_squared, i_team, i_rank = i_team_rating

            def od_reduce(od, q_team_ratings):
                omega, delta = od
                for q_team_rating in q_team_ratings:
                    q_mu, q_sigma_squared, q_team, q_rank = q_team_rating
                    ciq = 2 * math.sqrt(
                        i_sigma_squared + q_sigma_squared + self.TWO_BETA_SQUARED
                    )
                    delta_mu = (i_mu - q_mu) / ciq
                    sigma_squared_to_ciq = i_sigma_squared / ciq
                    gamma = self.gamma(ciq, len(self.team_ratings), *i_team_rating)

                    if q_rank > i_rank:
                        omega += sigma_squared_to_ciq * v(delta_mu, self.EPSILON / ciq)
                        delta += (
                            (gamma * sigma_squared_to_ciq)
                            / ciq
                            * w(delta_mu, self.EPSILON / ciq)
                        )
                    elif q_rank < i_rank:
                        omega += -sigma_squared_to_ciq * v(
                            -delta_mu, self.EPSILON / ciq
                        )
                        delta += (
                            (gamma * sigma_squared_to_ciq)
                            / ciq
                            * w(-delta_mu, self.EPSILON / ciq)
                        )
                    else:
                        omega += sigma_squared_to_ciq * vt(delta_mu, self.EPSILON / ciq)
                        delta += (
                            (gamma * sigma_squared_to_ciq)
                            / ciq
                            * wt(delta_mu, self.EPSILON / ciq)
                        )

                return omega, delta

            i_omega, i_delta = od_reduce([0, 0], i_adjacents)

            intermediate_result_per_team = []
            for j, j_players in enumerate(i_team):
                mu = j_players.mu
                sigma = j_players.sigma
                mu += (sigma ** 2) / i_sigma_squared * i_omega
                sigma *= math.sqrt(
                    max(1 - (sigma ** 2) / i_sigma_squared * i_delta, self.EPSILON)
                )
                intermediate_result_per_team.append([mu, sigma])
            return intermediate_result_per_team

        return list(
            map(
                lambda i: i_map(i[0], i[1]), zip(self.team_ratings, self.adjacent_teams)
            )
        )



import pytest

from openskill.rating import Rating
from openskill.util import team_rating

r = Rating()
team_1 = [r]
team_2 = [r, r]


def test_team_rating():
    # Aggregates all players in a team
    result = team_rating([team_1, team_2])
    assert result == [
        [25, 69.44444444444446, team_1, 0],
        [50, 138.8888888888889, team_2, 1],
    ]

    # 5 v 5
    result = team_rating([[r, r, r, r, r], [r, r, r, r, r]])
    assert result[0][0] == pytest.approx(125)
    assert result[1][0] == pytest.approx(125)
    assert result[0][1] == pytest.approx(347.22, 0.00001)
    assert result[1][1] == pytest.approx(347.22, 0.00001)



import pytest

from openskill.statistics import wt


def test_wt():
    assert wt(1, 2) == 0.3838582646421707
    assert wt(0, 2) == 0.2262586964500768
    assert wt(0, -1) == 1
    assert wt(0, 0) == 1.0
    assert wt(0, 10) == pytest.approx(0)



import math
from functools import reduce
from itertools import zip_longest
from typing import List, Optional

from openskill.constants import Constants
from openskill.rating import Rating


def score(q, i) -> float:
    if q < i:
        return 0.0

    if q > i:
        return 1.0

    return 0.5


def rankings(teams, rank: Optional[List[int]] = None):
    if rank:
        team_scores = [rank[i] or i for i, _ in enumerate(teams)]
    else:
        team_scores = [i for i, _ in enumerate(teams)]

    out_rank = {}
    s = 0
    for index, value in enumerate(team_scores):
        if index > 0:
            if team_scores[index - 1] < team_scores[index]:
                s = index
        out_rank[index] = s
    return list(out_rank.values())


def team_rating(game: List[List[Rating]], **options):
    if "rank" in options:
        rank = rankings(game, options["rank"])
    else:
        rank = rankings(game)

    result = []
    for index, team in enumerate(game):
        team_result = []
        mu_i = reduce(lambda x, y: x + y, map(lambda p: p.mu, team))
        sigma_squared = reduce(lambda x, y: x + y, map(lambda p: p.sigma ** 2, team))
        team_result.extend([mu_i, sigma_squared, team, rank[index]])
        result.append(team_result)
    return result


def ladder_pairs(ranks: List[int]):
    left = [None]
    left.extend(ranks[:-1])
    right = list(ranks[1:])
    right.append(None)
    zipped_lr = zip_longest(left, right)
    result = []
    for _left, _right in zipped_lr:
        if _left and _right:
            result.append([_left, _right])
        elif _left and not _right:
            result.append([_left])
        elif not _left and _right:
            result.append([_right])
        else:
            result.append([])
    return result


def util_c(team_ratings, **options):
    constants = Constants(**options)
    beta_squared = constants.BETA_SQUARED
    collective_team_sigma = 0
    for team in team_ratings:
        _team_mu, team_sigma_squared, _team, _rank = team
        collective_team_sigma += team_sigma_squared + beta_squared
    return math.sqrt(collective_team_sigma)


def util_sum_q(team_ratings, c):
    sum_q = {}
    for i, i_team in enumerate(team_ratings):
        team_i_mu, _team_i_sigma_squared, _i_team, i_rank = i_team
        temp = math.exp(team_i_mu / c)
        for q, q_team in enumerate(team_ratings):
            _team_q_mu, _team_q_sigma_squared, _q_team, q_rank = q_team
            if i_rank >= q_rank:
                if q in sum_q:
                    sum_q[q] += temp
                else:
                    sum_q[q] = temp
    return list(sum_q.values())


def util_a(team_ratings):
    result = list(
        map(
            lambda i: len(list(filter(lambda q: i[3] == q[3], team_ratings))),
            team_ratings,
        )
    )
    return result


def gamma(**options):
    if "gamma" in options:
        return options["gamma"]
    else:
        return (
            lambda c, _k, _mu, sigma_squared, _team, _q_rank: math.sqrt(sigma_squared)
            / c
        )


def transpose(xs):
    return [list(map(lambda r: r[i], xs)) for i, _ in enumerate(xs[0])]


def unwind(ranks, teams):
    if not ranks or not teams:
        return None, None

    def sorter(teams):
        unsorted_list = transpose([[ranks[i], [x, i]] for i, x in enumerate(teams)])
        sorted_list = [
            x
            for _, x in sorted(
                zip(unsorted_list[0], unsorted_list[1]), key=lambda pair: pair[0]
            )
        ]
        return [x for x, _ in sorted_list], [x for _, x in sorted_list]

    return sorter(teams) if isinstance(teams, list) else sorter



import math

from openskill.constants import Constants
from openskill.util import gamma, team_rating


class BradleyTerryFull:
    def __init__(self, game, **options):
        self.constants = Constants(**options)
        self.EPSILON = self.constants.EPSILON
        self.TWO_BETA_SQUARED = self.constants.TWO_BETA_SQUARED
        self.team_ratings = team_rating(game, **options)
        self.gamma = gamma(**options)

    def calculate(self):
        result = []
        for i, i_team_ratings in enumerate(self.team_ratings):
            omega = 0
            delta = 0
            i_mu, i_sigma_squared, i_team, i_rank = i_team_ratings

            for q, q_team_ratings in enumerate(self.team_ratings):
                q_mu, q_sigma_squared, q_team, q_rank = q_team_ratings

                if q == i:
                    continue

                ciq = math.sqrt(
                    i_sigma_squared + q_sigma_squared + self.TWO_BETA_SQUARED
                )
                piq = 1 / (1 + math.exp((q_mu - i_mu) / ciq))
                sigma_squared_to_ciq = i_sigma_squared / ciq

                s = 0
                if q_rank > i_rank:
                    s = 1
                elif q_rank == i_rank:
                    s = 0.5

                omega += sigma_squared_to_ciq * (s - piq)
                gamma = self.gamma(ciq, len(self.team_ratings), *i_team_ratings)
                delta += ((gamma * sigma_squared_to_ciq) / ciq) * piq * (1 - piq)

            intermediate_result_per_team = []
            for j, j_players in enumerate(i_team):
                mu = j_players.mu
                sigma = j_players.sigma
                mu += (sigma ** 2 / i_sigma_squared) * omega
                sigma *= math.sqrt(
                    max(1 - (sigma ** 2 / i_sigma_squared) * delta, self.EPSILON),
                )
                intermediate_result_per_team.append([mu, sigma])
            result.append(intermediate_result_per_team)
        return result



import random

from openskill.util import unwind


def test_unwind():
    # Zero Items
    src = []
    rank = []
    dst, de_rank = unwind(rank, src)
    assert dst is None
    assert de_rank is None

    # Accepts 1 Item
    src = ["a"]
    rank = [0]
    dst, de_rank = unwind(rank, src)
    assert dst == ["a"]
    assert de_rank == [0]

    # Accepts 2 Items
    src = ["b", "a"]
    rank = [1, 0]
    dst, de_rank = unwind(rank, src)
    assert dst == ["a", "b"]
    assert de_rank == [1, 0]

    # Accepts 3 Items
    src = ["b", "c", "a"]
    rank = [1, 2, 0]
    dst, de_rank = unwind(rank, src)
    assert dst == ["a", "b", "c"]
    assert de_rank == [2, 0, 1]

    # Accepts 4 Items
    src = ["b", "d", "c", "a"]
    rank = [1, 3, 2, 0]
    dst, de_rank = unwind(rank, src)
    assert dst == ["a", "b", "c", "d"]
    assert de_rank == [3, 0, 2, 1]

    # Can undo the ranking
    src = [random.random() for i in range(100)]
    random.shuffle(src)
    rank = [i for i in range(100)]
    trans, de_rank = unwind(rank, src)
    dst, de_de_rank = unwind(de_rank, trans)
    assert src == dst
    assert de_de_rank == rank

    # Allows ranks that are not zero-indexed integers
    src = ["a", "b", "c", "d", "e", "f"]
    rank = [0.28591, 0.42682, 0.35912, 0.21237, 0.60619, 0.47078]
    dst, de_rank = unwind(rank, src)
    assert dst == ["d", "a", "c", "b", "f", "e"]



from openskill.models.plackett_luce import PlackettLuce
from openskill.util import unwind


def rate(teams, **options):
    if "rank" in options:
        rank = options["rank"]
    else:
        if "score" in options:
            rank = list(map(lambda points: -points, options["score"]))
            options["rank"] = rank
        else:
            rank = None

    tenet = None
    if rank:
        ordered_teams, tenet = unwind(rank, teams)
        teams = ordered_teams
        rank = sorted(rank)
        options["rank"] = rank

    if "model" in options:
        model = options["model"](teams, **options)
    else:
        model = PlackettLuce(teams, **options)

    if rank and tenet:
        result = model.calculate()
        result, old_tenet = unwind(tenet, result)
        return result
    else:
        return model.calculate()



import pytest

from openskill.rating import Rating
from openskill.util import team_rating, util_c

r = Rating()
team_1 = [r]
team_2 = [r, r]


def test_util_c():
    # Compute Values
    team_ratings = team_rating([team_1, team_2])
    assert util_c(team_ratings) == pytest.approx(15.590239, 0.00000001)

    # Compute 5 v 5
    team_ratings = team_rating([[r, r, r, r, r], [r, r, r, r, r]])
    assert util_c(team_ratings) == pytest.approx(27.003, 0.00001)



from openskill.rating import Rating
from openskill.util import rankings

a = "a"
b = "b"
c = "c"
d = "d"
e = "e"

r = Rating()


def test_rankings():
    assert rankings([], None) == []
    assert rankings([], []) == []
    assert rankings([a], [a]) == [0]
    assert rankings([a, b, c, d], None) == [0, 1, 2, 3]
    assert rankings([a, b, c, d], [1, 2, 3, 4]) == [0, 1, 2, 3]
    assert rankings([a, b, c, d], [1, 1, 3, 4]) == [0, 0, 2, 3]
    assert rankings([a, b, c, d], [1, 2, 3, 3]) == [0, 1, 2, 2]
    assert rankings([a, b, c, d], [1, 2, 2, 4]) == [0, 1, 1, 3]
    assert rankings([a, b, c, d, e], [14, 32, 47, 47, 48]) == [0, 1, 2, 2, 4]



import pytest

from openskill import Rating, create_rating, rate
from openskill.models import ThurstonMostellerFull


def test_thurston_mosteller_full_series():
    p00 = Rating()
    p10 = Rating()
    p20 = Rating()
    p30 = Rating()
    p40 = Rating()

    result = rate(
        [[p00], [p10], [p20], [p30], [p40]],
        score=[9, 7, 7, 5, 5],
        epsilon=0.1,
        model=ThurstonMostellerFull,
    )
    [[p01], [p11], [p21], [p31], [p41]] = result
    p01 = create_rating(p01)
    p11 = create_rating(p11)
    p21 = create_rating(p21)
    p31 = create_rating(p31)
    p41 = create_rating(p41)

    p02 = p01
    p32 = p31
    result = rate(
        [[p41], [p21], [p11]], score=[9, 5, 5], epsilon=0.1, model=ThurstonMostellerFull
    )
    [[p42], [p22], [p12]] = result
    p42 = create_rating(p42)
    p22 = create_rating(p22)
    p12 = create_rating(p12)

    p43 = p42
    result = rate(
        [[p32], [p12], [p22], [p02]],
        score=[9, 9, 7, 7],
        epsilon=0.1,
        model=ThurstonMostellerFull,
    )
    [[p33], [p13], [p23], [p03]] = result
    p33 = create_rating(p33)
    p13 = create_rating(p13)
    p23 = create_rating(p23)
    p03 = create_rating(p03)

    assert p03.mu == pytest.approx(18.688153436)
    assert p03.sigma == pytest.approx(3.374349435)
    assert p13.mu == pytest.approx(27.015047867)
    assert p13.sigma == pytest.approx(3.250038289)
    assert p23.mu == pytest.approx(22.825795022)
    assert p23.sigma == pytest.approx(3.261730104)
    assert p33.mu == pytest.approx(27.281612237)
    assert p33.sigma == pytest.approx(3.387011242)
    assert p43.mu == pytest.approx(22.633338874)
    assert p43.sigma == pytest.approx(3.747505007)



from openskill.util import score


def test_score():
    assert score(2, 1) == 1.0
    assert score(34, 6) == 1.0
    assert score(1, 2) == 0.0
    assert score(3, 58) == 0.0
    assert score(1, 1) == 0.5



def z(**options) -> float:
    if "z" in options:
        return options["z"]
    else:
        return 3


def mu(**options) -> float:
    if "mu" in options:
        return options["mu"]
    else:
        return 25


def sigma(**options) -> float:
    if "sigma" in options:
        return options["sigma"]
    else:
        return mu(**options) / z(**options)


def epsilon(**options) -> float:
    if "epsilon" in options:
        return options["epsilon"]
    else:
        return 0.0001


def beta(**options) -> float:
    if "beta" in options:
        return options["beta"]
    else:
        return sigma(**options) / 2


def beta_squared(**options) -> float:
    if "beta_squared" in options:
        return options["beta_squared"]
    else:
        return beta(**options) ** 2


class Constants:
    def __init__(self, **options):
        self.EPSILON: float = epsilon(**options)
        self.TWO_BETA_SQUARED: float = 2 * beta_squared(**options)
        self.BETA_SQUARED: float = beta_squared(**options)
        self.Z: float = z(**options)



from openskill.rating import Rating
from openskill.util import team_rating, util_a

r = Rating()
team_1 = [r]
team_2 = [r, r]
team_3 = [r, r]
team_4 = [r]


def test_util_a():
    # Compute Values
    ratings = team_rating([team_1, team_2])
    assert util_a(ratings) == [1, 1]

    # 1 Team per Rank
    ratings = team_rating([team_1, team_2, team_3, team_4])
    assert util_a(ratings) == [1, 1, 1, 1]

    # Count how many share the rank
    ratings = team_rating([team_1, team_2, team_3, team_4], rank=[1, 1, 1, 4])
    assert util_a(ratings) == [3, 3, 3, 1]



import math

from openskill.constants import Constants
from openskill.statistics import v, vt, w, wt
from openskill.util import gamma, team_rating


class ThurstonMostellerFull:
    def __init__(self, game, **options):
        self.constants = Constants(**options)
        self.EPSILON = self.constants.EPSILON
        self.TWO_BETA_SQUARED = self.constants.TWO_BETA_SQUARED
        self.team_ratings = team_rating(game, **options)
        self.gamma = gamma(**options)

    def calculate(self):
        result = []
        for i, i_team_ratings in enumerate(self.team_ratings):
            omega = 0
            delta = 0
            i_mu, i_sigma_squared, i_team, i_rank = i_team_ratings

            for q, q_team_ratings in enumerate(self.team_ratings):
                q_mu, q_sigma_squared, q_team, q_rank = q_team_ratings

                if q == i:
                    continue

                ciq = math.sqrt(
                    i_sigma_squared + q_sigma_squared + self.TWO_BETA_SQUARED
                )
                delta_mu = (i_mu - q_mu) / ciq
                sigma_squared_to_ciq = i_sigma_squared / ciq
                gamma = self.gamma(ciq, len(self.team_ratings), *i_team_ratings)

                if q_rank > i_rank:
                    omega += sigma_squared_to_ciq * v(delta_mu, self.EPSILON / ciq)
                    delta += (
                        gamma
                        * sigma_squared_to_ciq
                        / ciq
                        * w(delta_mu, self.EPSILON / ciq)
                    )
                elif q_rank < i_rank:
                    omega += -sigma_squared_to_ciq * v(-delta_mu, self.EPSILON / ciq)
                    delta += (
                        gamma
                        * sigma_squared_to_ciq
                        / ciq
                        * w(-delta_mu, self.EPSILON / ciq)
                    )
                else:
                    omega += sigma_squared_to_ciq * vt(delta_mu, self.EPSILON / ciq)
                    delta += (
                        gamma
                        * sigma_squared_to_ciq
                        / ciq
                        * wt(delta_mu, self.EPSILON / ciq)
                    )

            intermediate_result_per_team = []
            for j, j_players in enumerate(i_team):
                mu = j_players.mu
                sigma = j_players.sigma
                mu += (sigma ** 2 / i_sigma_squared) * omega
                sigma *= math.sqrt(
                    max(1 - (sigma ** 2 / i_sigma_squared) * delta, self.EPSILON),
                )
                intermediate_result_per_team.append([mu, sigma])
            result.append(intermediate_result_per_team)
        return result



from typing import List, Optional, Union

from openskill.constants import mu as default_mu
from openskill.constants import sigma as default_sigma


class Rating:
    def __init__(
        self, mu: Optional[float] = None, sigma: Optional[float] = None, **options
    ):
        # Calculate Mu and Sigma
        self.mu = mu if mu else default_mu(**options)
        self.sigma = sigma if sigma else default_sigma(**options)

    def __repr__(self):
        return f"Rating(mu={self.mu}, sigma={self.sigma})"

    def __eq__(self, other):
        if len(other) == 2:
            if self.mu == other[0] and self.sigma == other[1]:
                return True
            else:
                return False


def create_rating(rating_list: List[Union[int, float]]):
    return Rating(mu=rating_list[0], sigma=rating_list[1])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openskill', 'openskill.models']

package_data = \
{'': ['*']}

install_requires = \
['scipy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'openskill',
    'version': '0.1.0a2',
    'description': 'Python implementation of Weng-Lin Bayesian ranking, a better, license-free alternative to TrueSkill.',
    'long_description': "![](https://i.imgur.com/tSTFzZY.gif)\n\n\n[![Tests](https://github.com/OpenDebates/openskill.py/actions/workflows/main.yml/badge.svg)](https://github.com/OpenDebates/openskill.py/actions/workflows/main.yml) [![codecov](https://codecov.io/gh/OpenDebates/openskill.py/branch/main/graph/badge.svg?token=Ep07QEelsi)](https://codecov.io/gh/OpenDebates/openskill.py) ![PyPI - Downloads](https://img.shields.io/pypi/dm/openskill)\n\nPython implementation of Weng-Lin Bayesian ranking, a better, license-free alternative to TrueSkill\n\nThis is a port of the amazing [openskill.js](https://github.com/philihp/openskill.js) package.\n\n## Installation\n```shell\npip install openskill\n```\n\n## Usage\n```python\n>>> from openskill import Rating, rate\n>>> a1 = Rating()\n>>> a1\nRating(mu=25, sigma=8.333333333333334)\n>>> a2 = Rating(mu=32.444, sigma=5.123)\n>>> a2\nRating(mu=32.444, sigma=5.123)\n>>> b1 = Rating(43.381, 2.421)\n>>> b1\nRating(mu=43.381, sigma=2.421)\n>>> b2 = Rating(mu=25.188, sigma=6.211)\n>>> b2\nRating(mu=25.188, sigma=6.211)\n```\n\nIf `a1` and `a2` are on a team, and wins against a team of `b1` and `b2`, send this into rate:\n\n```python\n>>> [[x1, x2], [y1, y2]] = rate([[a1, a2], [b1, b2]])\n>>> x1, x2, y1, y2\n([28.669648436582808, 8.071520788025197], [33.83086971107981, 5.062772998705765], [43.071274808241974, 2.4166900452721256], [23.149503312339064, 6.1378606973362135])\n```\n\nYou can also create `Rating` objects by importing `create_rating`:\n\n```python\n>>> from openskill import create_rating\n>>> x1 = create_rating(x1)\n>>> x1\nRating(mu=28.669648436582808, sigma=8.071520788025197)\n```\n\n## Ranks\nWhen displaying a rating, or sorting a list of ratings, you can use `ordinal`:\n\n```python\n>>> from openskill import ordinal\n>>> ordinal(mu=43.07, sigma=2.42)\n35.81\n```\n\nBy default, this returns `mu - 3 * sigma`, showing a rating for which there's a [99.7%](https://en.wikipedia.org/wiki/68%E2%80%9395%E2%80%9399.7_rule) likelihood the player's true rating is higher, so with early games, a player's ordinal rating will usually go up and could go up even if that player loses.\n\n## Artificial Ranks\nIf your teams are listed in one order but your ranking is in a different order, for convenience you can specify a ranks option, such as:\n\n```python\n>>> a1 = b1 = c1 = d1 = Rating()\n>>> result = [[a2], [b2], [c2], [d2]] = rate([[a1], [b1], [c1], [d1]], rank=[4, 1, 3, 2])\n>>> result\n[[[20.96265504062538, 8.083731307186588]], [[27.795084971874736, 8.263160757613477]], [[24.68943500312503, 8.083731307186588]], [[26.552824984374855, 8.179213704945203]]]\n```\n\nIt's assumed that the lower ranks are better (wins), while higher ranks are worse (losses). You can provide a score instead, where lower is worse and higher is better. These can just be raw scores from the game, if you want.\n\nTies should have either equivalent rank or score.\n\n```python\n>>> a1 = b1 = c1 = d1 = Rating()\n>>> result = [[a2], [b2], [c2], [d2]] = rate([[a1], [b1], [c1], [d1]], score=[37, 19, 37, 42])\n[[[24.68943500312503, 8.179213704945203]], [[22.826045021875203, 8.179213704945203]], [[24.68943500312503, 8.179213704945203]], [[27.795084971874736, 8.263160757613477]]]\n```\n\n\n## Choosing Models\n\nThe default model is `PlackettLuce`. You can import alternate models from `openskill.models` like so:\n\n```python\n>>> from openskill.models import BradelyTerryFull\n>>> a1 = b1 = c1 = d1 = Rating()\n>>> rate([[a1], [b1], [c1], [d1]], rank=[4, 1, 3, 2], model=BradleyTerryFull)\n[[[17.09430584957905, 7.5012190693964005]], [[32.90569415042095, 7.5012190693964005]], [[22.36476861652635, 7.5012190693964005]], [[27.63523138347365, 7.5012190693964005]]]\n```\n\n### Available Models\n- `BradleyTerryFull`: Full Pairing for Bradley-Terry\n- `BradleyTerryPart`: Partial Pairing for Bradely-Terry\n- `PlackettLuce`: Generalized Bradley-Terry\n- `ThurstonMostellerFull`: Full Pairing for Thurston-Mosteller\n- `ThurstonMostellerPart`: Partial Pairing for Thurston-Mosteller\n\n### Which Model Do I Want?\n\n- Bradley-Terry rating models follow a logistic distribution over a player's skill, similar to Glicko.\n- Thurstone-Mosteller rating models follow a gaussian distribution, similar to TrueSkill. Gaussian CDF/PDF functions differ in implementation from system to system (they're all just chebyshev approximations anyway). The accuracy of this model isn't usually as great either, but tuning this with an alternative gamma function can improve the accuracy if you really want to get into it.\n- Full pairing should have more accurate ratings over partial pairing, however in high k games (like a 100+ person marathon race), Bradley-Terry and Thurston-Mosteller models need to do a calculation of joint probability which involves is a k-1 dimensional integration, which is computationally expensive. Use partial pairing in this case, where players only change based on their neighbors.\n- Plackett-Luce (**default**) is a generalized Bradley-Terry model for k ≥ 3 teams. It scales best.\n\n\n## Implementations in other Languages\n- [Javascript](https://github.com/philihp/openskill.js)\n- [Elixir](https://github.com/philihp/openskill.ex)\n- [Kotlin](https://github.com/brezinajn/openskill)\n\n",
    'author': 'Taven',
    'author_email': 'taven@outlook.in',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/OpenDebates/openskill.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import math

from openskill.constants import Constants
from openskill.util import gamma, team_rating, util_a, util_c, util_sum_q


class PlackettLuce:
    def __init__(self, game, **options):
        self.constants = Constants(**options)
        self.EPSILON = self.constants.EPSILON
        self.team_ratings = team_rating(game, **options)
        self.c = util_c(self.team_ratings, **options)
        self.gamma = gamma(**options)
        self.sum_q = util_sum_q(self.team_ratings, self.c)
        self.a = util_a(self.team_ratings)

    def calculate(self):
        result = []
        for i, i_team_ratings in enumerate(self.team_ratings):
            omega = 0
            delta = 0
            i_mu, i_sigma_squared, i_team, i_rank = i_team_ratings
            i_mu_over_ce = math.exp(i_mu / self.c)

            for q, q_team_ratings in enumerate(self.team_ratings):
                q_mu, q_sigma_squared, q_team, q_rank = q_team_ratings
                i_mu_over_ce_over_sum_q = i_mu_over_ce / self.sum_q[q]
                if q_rank <= i_rank:
                    delta += (
                        i_mu_over_ce_over_sum_q
                        * (1 - i_mu_over_ce_over_sum_q)
                        / self.a[q]
                    )
                    if q == i:
                        omega += (1 - i_mu_over_ce_over_sum_q) / self.a[q]
                    else:
                        omega -= i_mu_over_ce_over_sum_q / self.a[q]

            omega *= i_sigma_squared / self.c
            delta *= i_sigma_squared / self.c ** 2

            gamma = self.gamma(self.c, len(self.team_ratings), *i_team_ratings)
            delta *= gamma

            intermediate_result_per_team = []
            for j, j_players in enumerate(i_team):
                mu = j_players.mu
                sigma = j_players.sigma
                mu += (sigma ** 2 / i_sigma_squared) * omega
                sigma *= math.sqrt(
                    max(1 - (sigma ** 2 / i_sigma_squared) * delta, self.EPSILON),
                )
                intermediate_result_per_team.append([mu, sigma])
            result.append(intermediate_result_per_team)
        return result



import sys

import scipy.stats

normal = scipy.stats.norm(0, 1)


def phi_major(x):
    return normal.cdf(x)


def phi_minor(x):
    return normal.pdf(x)


def v(x, t):
    xt = x - t
    denom = phi_major(xt)
    return -xt if (denom < sys.float_info.epsilon) else phi_minor(xt) / denom


def w(x, t):
    xt = x - t
    denom = phi_major(xt)
    if denom < sys.float_info.epsilon:
        return 1 if (x < 0) else 0
    return v(x, t) * (v(x, t) + xt)


def vt(x, t):
    xx = abs(x)
    b = phi_major(t - xx) - phi_major(-t - xx)
    if b < 1e-5:
        if x < 0:
            return -x - t
        return -x + t
    a = phi_minor(-t - xx) - phi_minor(t - xx)
    return (-a if x < 0 else a) / b


def wt(x, t):
    xx = abs(x)
    b = phi_major(t - xx) - phi_major(-t - xx)
    if b < sys.float_info.epsilon:
        return 1.0
    return ((t - xx) * phi_minor(t - xx) + (t + xx) * phi_minor(-t - xx)) / b + vt(
        x, t
    ) * vt(x, t)
